//
//  Cube.java
//
//  Created by Thomas Jannaud on 09/08/08.
//  Copyright 2008 Jannaud's Technical Company. All rights reserved.
//

// il y a 4 parties :
// 1) les valeurs "statiques", initialisées par "initialiser()" : table des couleurs, ...
// 2) la construction du cube. On donne essentiellement 3 manières de faire
// 3) calcul des 4 coordonnées 3D d'une face ou bien d'un "sticker" étant donné la face à laquelle il appartient, et n° ligne et colonne
// 4) les fonctions de dessin. dessiner_rectangle() ne sert qu'à faire les traits noirs séparateurs.

import java.awt.*;
import javax.swing.*;
import java.util.*;

public class Cube {
	private Color[] lecube;
	public static HashMap<Character, Color> couleur = new HashMap<Character, Color>();
	private static Point3D[] liste_coordonnees;
	private static Color[] lescouleurs;
	
	public static void initialiser() {
		couleur.put('L', Color.blue); // ATTENTION pour le bleu
		couleur.put('R', Color.red);
		couleur.put('B', Color.white);
		couleur.put('O', new Color(255, 86, 0));
		couleur.put('J', Color.yellow);
		couleur.put('V', Color.green);
		couleur.put('G', Color.gray);
		couleur.put('N', Color.black);
		liste_coordonnees = new Point3D[8];
		liste_coordonnees[0] = new Point3D(-.5, .5, .5);
		liste_coordonnees[1] = new Point3D(.5, .5, .5);
		liste_coordonnees[2] = new Point3D(-.5, -.5, .5);
		liste_coordonnees[3] = new Point3D(.5, -.5, .5);
		liste_coordonnees[4] = new Point3D(-.5, .5, -.5);
		liste_coordonnees[5] = new Point3D(.5, .5, -.5);
		liste_coordonnees[6] = new Point3D(-.5, -.5, -.5);
		liste_coordonnees[7] = new Point3D(.5, -.5, -.5);
		lescouleurs = new Color[6];
		lescouleurs[0] = Color.blue;
		lescouleurs[1] = Color.red;
		lescouleurs[2] = Color.white;
		lescouleurs[3] = new Color(255, 86, 0);
		lescouleurs[4] = Color.yellow;
		lescouleurs[5] = Color.green;
	}
	Cube() {
		lecube = new Color[54];
		for(int i = 0; i < 54; i++)
			lecube[i] = Color.gray;
	}
	Cube(String s) {
		/* premier caractère : 0, 1 ou 2, suivant que l'on veuille :
		 0 : un cube "résolu" en entier
		 1 : un cube que l'on décrit face par face
		 face 0 : U, face 1 : F, 2 : L, 3 : B, 4 : R, 5 : D  (B : back).
		 en regardant une face devant nous on décrit de gauche à droite et de haut en bas. (avec U = bleu, F = rouge, R = jaune comme référence)
		 2 : un cube que l'on décrit comme ceci : 
		 de 0 à 8 : face du haut
		 de 9 à 20 : "bande" du haut, partant de UFR (partie F) puis tournant et revenant à UFR (partie R)
		 21 à 32 : idem mais 2eme bande
		 33 à 44 : idem mais 3eme bande
		 45 à 53 : face du bas
		 */
		lecube = new Color[54];
		if (s.charAt(0) == '0') {
			for(int i = 0; i < 6; i++)
				for(int j = 0; j < 9; j++)
					lecube[9 * i + j] = lescouleurs[i];
		}
		else if (s.charAt(0) == '1') {
			for(int i = 0; i < (s.length() - 1); i++)
				lecube[i] = couleur.get(s.charAt(i + 1));
			for(int i = s.length() - 1; i < 54; i++)
				lecube[i] = Color.gray;
		}
		else {
			for(int i = 0; i < (s.length() - 1); i++)
				lecube[f(i)] = couleur.get(s.charAt(i + 1));
			for(int i = s.length() - 1; i < 54; i++)
				lecube[f(i)] = Color.gray;
		}
	}
	private int f(int i) { // fonction qui à la numérotation "secondaire" (par bandes) fait correspondre la première numérotation
		if (i <= 8 || i >= 45)
			return i;
		if (i >= 9 && i <= 20)
			return (9 + 9 * ((i - 9) / 3) + 2 - (i % 3));
		if (i >= 21 && i <= 32)
			return (9 + 9 * ((i - 21) / 3) + 5 - (i % 3));
		if (i >= 33 && i <= 44)
			return (9 + 9 * ((i - 33) / 3) + 8 - (i % 3));
		throw new Error("hors de la numérotation");
	}
	
	private Point3D[] lecoin(int a, int b, int c, int d) {
		Point3D[] retour = {liste_coordonnees[a], liste_coordonnees[b], liste_coordonnees[c], liste_coordonnees[d]};
		return retour;
	}
	private Point3D[] get_coordonnees(int nface) {
		if (nface == 0) // bleu = U
			return lecoin(2, 0, 1, 3);
		if (nface == 1) // rouge = F
			return lecoin(3, 1, 5, 7);
		if (nface == 2) // blanc = L
			return lecoin(2, 3, 7, 6);
		if (nface == 3) // orange = B
			return lecoin(0, 2, 6, 4);
		if (nface == 4) // jaune = R
			return lecoin(1, 0, 4, 5);
		if (nface == 5) // vert = D
			return lecoin(7, 5, 4, 6);
		throw new Error("nface n'est pas dans les bornes");
	}
	private Point3D[] get_coordonnees(int nface, int ligne, int colonne) {
		Point3D[] p = get_coordonnees(nface);
		/*		System.out.println("coordonnées de la face : ");
		 for(int i = 0; i < 4; i++)
		 System.out.println(p[i]);*/
		Point3D[] retour = new Point3D[4];
		double ratiox = ((double) (colonne)) / 3.0, ratioy = ((double) (ligne)) / 3.0;
		retour[0] = new Point3D(p[0].x + (p[1].x - p[0].x) * ratiox + (p[3].x - p[0].x) * ratioy, p[0].y + (p[1].y - p[0].y) * ratiox + (p[3].y - p[0].y) * ratioy, p[0].z + (p[1].z - p[0].z) * ratiox + (p[3].z - p[0].z) * ratioy);
		retour[1] = new Point3D(retour[0].x + (p[1].x - p[0].x) / 3.0, retour[0].y + (p[1].y - p[0].y) / 3.0, retour[0].z + (p[1].z - p[0].z) / 3.0);
		retour[2] = new Point3D(retour[0].x + (p[1].x - p[0].x) / 3.0 + (p[3].x - p[0].x) / 3.0, retour[0].y + (p[1].y - p[0].y) / 3.0 + (p[3].y - p[0].y) / 3.0, retour[0].z + (p[1].z - p[0].z) / 3.0 + (p[3].z - p[0].z) / 3.0);
		retour[3] = new Point3D(retour[0].x + (p[3].x - p[0].x) / 3.0, retour[0].y + (p[3].y - p[0].y) / 3.0, retour[0].z + (p[3].z - p[0].z) / 3.0);
		return retour;	
	}
	
	public void dessiner(Graphics g, Viewer camera) {
		g.clearRect(0, 0, 500, 500);
		boolean[] isVisible = camera.faces_visibles();
		for(int i = 0; i < 6; i++)
			if (isVisible[i])
				dessine_face(i, g, camera);		
	}
	private void dessine_face(int nface, Graphics g, Viewer camera) {
		Point3D[][] stickers = new Point3D[9][];
		
		for(int i = 0; i < 3; i++)
			for(int j = 0; j < 3; j++) {
				Point3D[] sticker = get_coordonnees(nface, i, j);
				stickers[3*i+j] = sticker;
				dessiner(sticker, lecube[9 * nface + 3 * i + j] , g, camera);
			}
		
		// dessine les traits noirs entre les stickers
		dessiner_rectangle(stickers[0][0], stickers[2][1], stickers[8][2], stickers[6][3], g, camera);
		dessiner_rectangle(stickers[1][0], stickers[1][1], stickers[7][2], stickers[7][3], g, camera);
		dessiner_rectangle(stickers[3][0], stickers[3][3], stickers[5][2], stickers[5][1], g, camera);
	}
	private void dessiner(Point3D[] coins, Color col, Graphics g, Viewer camera) {
	// on donne 4 points dans l'espace (un carré en fait), et cette fonction le dessine
	// en faisant appel aux fonctions de projections 3D -> 2D de la caméra

		Point2D[] proj = new Point2D[4];
		for (int i = 0; i < 4; i++)
			proj[i] = camera.projeter(coins[i]);
		int[][] liste = camera.coordonnees(proj);

		g.setColor(col);
		g.fillPolygon(liste[0], liste[1] , 4);
	}
	private void dessiner_rectangle(Point3D a, Point3D b, Point3D c, Point3D d, Graphics g, Viewer camera) {
		Point2D[] proj = new Point2D[4];
		proj[0] = camera.projeter(a);
		proj[1] = camera.projeter(b);
		proj[2] = camera.projeter(c);
		proj[3] = camera.projeter(d);
		g.setColor(Color.black);
		int[][] liste = camera.coordonnees(proj);
		g.drawPolygon(liste[0], liste[1] , 4);
	}
}
